;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

(defparameter *listing-browser-size*
  (xy 300 (+ 70 (* 10 (font-height *listing-font*)) 16)))

(defclass listing-browser (browser)
  ()
  (:default-initargs 
    :view-size *listing-browser-size*
    :color-p t
    :window-do-first-click t))

(defun listing-browser? (object)
  (typep object 'listing-browser))

(defmethod make-browser ((class (eql 'listing-browser)) object &rest args)
  (let ((browser
         (apply
          #'make-instance 'listing-browser 
          :object object
          :window-title (object-namestring object t)
          :view-subviews
          (list
           (make-listing-control-view 
            :view-size (xy (xy-x *listing-browser-size*) 70))
           (make-instance 'ccl:pop-up-menu 
             :auto-update-default nil
             :view-position (xy 8 30) ; 40
             :view-font *capella-font*
             :menu-items
             (loop for o in (object-superiors object)
                   collect
                   (let ((o o))
                     (make-instance 'ccl:menu-item
                       :menu-item-title (object-namestring o)
                       :menu-item-action #'(lambda () (browser o))))))
           (make-instance 'check-box :title "Display Time"
                          :view-position (xy 8 50) 
                          :action 
                          #'(lambda (c &aux l)
                              (setf l (browser-listing (view-browser c)))
                              (setf (listing-view-show-time? l) 
                                    (toggle-selected? c))
                              (set-listing-display-x l)
                              (ccl:invalidate-view l)))
           
           (make-instance 'selection-view
             :view-position (xy 110 27) :view-size (xy 0 35)) 
           (make-listing-view object
                              :view-position (xy 0 70)
                              :view-size 
                              (xy 300 (* 10 (font-height *listing-font*)))))
          args)))
    (update-browser browser)
    (set-browser-title browser)
    (set-window-grow browser *listing-browser-size*)
    browser))

(defmethod ccl:window-can-do-operation ((window listing-browser)
                                        op &optional item)
  (declare (ignore item))
  (find op '(ccl:cut ccl:copy ccl:paste ccl:clear ccl:select-all) 
        :test #'eq)) ;search ccl:window-save

(defmethod ccl:window-size-parts ((browser listing-browser))
  (when (> (length (ccl:view-subviews browser)) 0)
    (let* ((size (ccl:view-size browser))
           (width (xy-x size)))
      ;; put selection view 1/3 across window sized to window-pos-border
      (let ((v (ccl:view-named ':selection-view browser)))
        (ccl:set-view-size v (- width (xy-x (ccl:view-position v)) 10) 
                           (view-height v)))
      ;; spread control menu across window.
      (let ((v (ccl:view-named ':control-view browser)))
        (ccl:set-view-size v width (view-height v)))
      (let ((v (browser-listing browser)))
        (ccl:set-view-size v 
                           (xy (- width 15) ; -15 needed for scroll-bars
                               (- (xy-y size) 
                                  15
                                  (xy-y (ccl:view-position v)))))
        (update-sequence-visible v)))))

(defmethod set-browser-title ((browser listing-browser) &optional object)
  (ccl:set-window-title
   browser
   (format nil "~A Listing" 
           (object-namestring (or object (object-view-object browser))))))

(defmethod update-browser ((browser listing-browser))
  (update-listing-controls browser)
  (update-listing-status browser))

(defun browser-listing (browser)
  (ccl:view-named ':listing-view browser))

(defmethod update-window-for-changed-object ((browser listing-browser)
                                              object)
  (if (eq (object-view-object browser) object)
    (set-browser-title browser object)
    (let* ((listing (browser-listing browser))
           (position (position object (table-sequence listing))))
      (when position
        (clear-display (listing-view-displays listing) object)
        (redraw-cell listing 0 position)))))

(defun object-superiors (object &optional (include? t))
  (let ((levels (if (typep object 'id-mixin)
                  (if (systemp object) 
                    (if (eq object .toc.) nil (list .toc.))
                    (list .top-level.))
                  (let ((sup (the-container object nil)))
                    (if sup (list .top-level.) nil)))))
    (loop with o = object
          for s = (object-containers o)
          while (and s (= (length s) 1))
          do (progn (setf o (car s)) (push o levels)))
    (when include? (push object levels))
    levels))
;;;
;;; browser window and selection operators call the selection view.
;;;

(defmethod ccl:select-all ((browser listing-browser))
  (let ((handler (ccl:current-key-handler browser)))
    (if handler
      (ccl:select-all handler)
      (let ((listing (browser-listing browser)))
        (ccl:select-all listing)
        (set-selection-typein browser listing)
        (update-browser browser)))))

(defmethod ccl:clear ((browser listing-browser))
  (let ((handler (ccl:current-key-handler browser)))
    (if handler
      (ccl:clear handler)
      (let ((listing (browser-listing browser)))
        (ccl:clear listing)
        (set-selection-typein browser listing)
        (update-browser browser)))))

(defmethod ccl:copy ((browser listing-browser))
  (let ((handler (ccl:current-key-handler browser)))
    (if handler
      (ccl:copy handler)
      (ccl:copy (browser-listing browser)))))

(defmethod ccl:paste ((browser listing-browser))
  (let ((handler (ccl:current-key-handler browser)))
    (if handler
      (ccl:paste handler)
      (progn  (ccl:paste (browser-listing browser))
              (update-browser browser)))))

(defmethod ccl:cut ((browser listing-browser))
  (let ((handler (ccl:current-key-handler browser)))
    (if handler
      (ccl:cut handler)
      (progn (ccl:cut (browser-listing browser))
             (update-browser browser)))))

(defmethod single-selection? ((browser listing-browser))
  (single-selection? (browser-listing browser)))

(defmethod disjoint-selection? ((browser listing-browser))
  (disjoint-selection? (browser-listing browser)))

(defmethod contiguous-selection? ((browser listing-browser))
  (contiguous-selection? (browser-listing browser)))

(defmethod any-selection? ((browser listing-browser))
  (any-selection? (browser-listing browser)))

(defmethod no-selection? ((browser listing-browser))
  (no-selection? (browser-listing browser)))

(defmethod selection-class? ((browser listing-browser) &optional max typep)
  (selection-class? (browser-listing browser) max typep))

(defmethod first-selection ((browser listing-browser))
  (first-selection (browser-listing browser)))

(defmethod last-selection ((browser listing-browser))
  (last-selection (browser-listing browser)))

;;;
;;; the workhorse.
;;;

(defclass listing-view (sequence-view object-view-mixin)
  ((displays :initarg :listing-displays :accessor listing-view-displays
             :initform (make-hash-table))
   (display-x :accessor listing-view-display-x)
   (show-time? :initform nil :accessor listing-view-show-time?)
   (reference :initform nil :accessor listing-view-reference))
  (:default-initargs
    :view-font *listing-font*
    :view-nick-name ':listing-view))

(defmethod set-table-sequence :after ((view listing-view) sequence)
  (declare (ignore sequence))
  (set-listing-display-x view))

(defun set-listing-display-x (view)
  (setf (slot-value view 'display-x)
        (ccl:string-width 
         (if (slot-value view 'show-time?)
           (format nil "0000.000 ~D.  " (sequence-length view))
           (format nil "~D.  " (sequence-length view)))                                 
         *listing-font*)))

(defun update-listing-times (listing)
  (setf (slot-value listing 'display-x)
        ())
  (ccl:invalidate-view listing))

(defun make-listing-view (object &rest args)
  (let ((entries (container-objects object)))
    (apply #'make-instance 'listing-view
           :object object
           :view-size (xy 350 (* 10 (font-height *listing-font*)))
           :table-sequence entries
           args)))

(defmethod initialize-instance :after ((view listing-view) &rest args)
  (declare (ignore args))
  (setf (slot-value view 'displays)
        (make-hash-table :size (sequence-length view)))
  (ccl:add-subviews view (slot-value view 'ccl::v-scroller)
                    (slot-value view 'ccl::h-scroller)))

(defmethod ccl:view-click-event-handler :after ((listing listing-view) where)
  (declare (ignore where))
  (let ((browser (view-browser listing))) 
    (set-selection-typein browser listing)
    (when (and (= (sequence-nselected listing) 1)
               (ccl:double-click-p))
      (map-selected-objects #'browser listing))
    (update-browser browser)
    (ccl:set-current-key-handler browser nil)))

(defun clear-display (displays object)
  ;; clear current display cache.
  (setf (gethash object displays) nil))

(defun draw-changed-displays (listing)
  ;; only redisplay visible cells with no cache
  (let ((beg (sequence-first-visible listing))
        (end (sequence-last-visible listing))
        (tbl (listing-view-displays listing)))
    (loop for o in (nthcdr beg (table-sequence listing))
          for i from beg to end
          unless (gethash o tbl)
          do (redraw-cell listing 0 i))))

(defparameter %posbuf (make-string 4)) ; nonconsing position printing

(defmethod draw-cell-contents ((view listing-view) col &optional row)
  (multiple-value-setq (col row) (canonicalize-point-args :values col row))
  (let ((entry (cell-contents view col row))
        (display (slot-value view 'displays))
        (xpos (slot-value view 'display-x))
        (index (1+ row))
        (color ccl:*black-color*)
        (ypos (xy-y (ccl:pen-position view)))
        flags string )
    (if (logtest (setf flags (object-flags entry)) +deleted+)
      (setf color ccl:*light-gray-color*)
      (if (logtest flags +hidden+)
        (setf color ccl:*dark-gray-color*)))
    ;; print strings cached for rapid redisplay.
    (setf string (or (gethash entry display)
                     (setf (gethash entry display)
                           (prin1-to-string entry))))
    ;; position formatting reuses string.
    (ccl:with-fore-color color
      (ccl:with-pstrs ((s1 "  ") ;(fi2s index %posbuf 4)
                       (s2 string)
                       (s3 ""))
        (when (slot-value view 'show-time?)
          (let ((time (if (slot-boundp entry 'time) 
                        (slot-value entry 'time))))
            (when time
              (#_moveto 3 ypos)
              (multiple-value-bind (n r) (floor time)
                (#_NumToString n s3)
                (dotimes (i (- 4 (ccl:%get-byte s3)))
                  (#_drawchar #\space))
                (#_drawstring s3)
                (#_drawchar #\.)
                (#_numtostring (if (= r 0.0) 0 (floor r .001)) s3)
                (#_drawstring s3)
                (dotimes (i (- 3 (ccl:%get-byte s3)))
                  (#_drawchar #\0))))))
        (let ((x0 (- xpos (#_stringwidth s1))))
          (#_moveto x0 ypos)
          (#_drawchar #\.)
          (#_moveto x0 ypos)
          (#_NumToString index s1)
          (#_move (- (#_stringwidth s1)) 0)
          (#_DrawString s1)
          (#_MoveTo xpos ypos)
          (#_DrawString s2))))))

(defun fi2s (i s &optional (w 4))
  ;; format integer to string with space padding
  (loop for m = (floor i 10)
        do
        (setf (elt s (decf w)) (code-char (+ 48 (- i (* 10 m)))))
        (setf i m)
        until (= i 0)
        finally (loop for i below w
                      do (setf (elt s i) #\Space)))
  s)

;;; 
;;; add stella methods for reference mapping
;;;

(defmethod object-count ((listing listing-view))
  (sequence-length listing))

(defmethod subobjectsp ((listing listing-view))
  (> (sequence-length listing) 0))

(defmethod container-objects ((view listing-view))
  (table-sequence view))

;;;
;;; listing view selection functions.
;;;

(defmethod first-selection ((view listing-view))
  (first-selected-cell view))

(defmethod last-selection ((view listing-view))
  (first-selected-cell view t))

(defmethod single-selection? ((view listing-view))
  (= (sequence-nselected view) 1))

(defmethod disjoint-selection? ((view listing-view))
  ;; true if there are selection but no selection region
  (let ((first (first-selection view)))
    (and first
         (loop for i from first below (last-selection view)
               thereis (not (cell-selected-p view 0 i))))))

(defmethod contiguous-selection? ((listing listing-view))
  ;; true if all selections are in a single range
  (let ((lb (first-selection listing)))
    (and lb
         (let ((ub
                (loop for i from lb below (sequence-length listing)
                      while (cell-selected-p listing 0 i)
                      finally return i)))
           (loop for i from ub below (sequence-length listing)
                 never (cell-selected-p listing 0 i))))))

(defmethod any-selection? ((listing listing-view))
  (> (sequence-nselected listing) 0))

(defmethod no-selection? ((listing listing-view))
  (not (any-selection? listing)))

(defmethod selection-class? ((listing listing-view) &optional max typep)
  ;; return class of objects selected if all are the same.
  (let ((type typep)
        (count 0))
    (block checking
      (map-selected-objects
       #'(lambda (o &aux (c (class-of o)))
           (if type
             (unless (or (eq type c) (and typep (typep o type)))
               (return-from selection-class? nil))
             (setf type c))
           (and max (= (incf count) max) (return-from checking nil)))
       listing))
    type))

(defvar %selstr
  (make-array 256 :adjustable t :element-type 'character
              :fill-pointer t))

(defun selections-to-string (view)
  ;; decompile selected listing entries into a reference string
  (let ((n (sequence-nselected view))
        (l (sequence-length view))
        (i 0)
        (j 0))
    (declare (integer n l i j))
    (setf (fill-pointer %selstr) 0)
    (unless (= n 0)
      (if (= n l)
        (vector-push #\* (the string %selstr))
        (flet ((itos (i s)
                 (ccl:with-pstrs ((p ""))
                   (#_NumToString i p)
                   (dotimes (j (ccl:%get-byte p))
                     (vector-push-extend
                      (the character
                        (code-char (ccl:%get-byte p (+ 1 j))))
                      (the string s))))))
          (loop while (< i l)
                do
                (setf j (1+ i))
                (if (cell-selected-p view 0 i)
                  (progn
                    (vector-push #\, (the string %selstr))
                    (itos (1+ i) %selstr)
                    (loop while (and (< j l)
                                     (cell-selected-p view 0 j))
                          do (incf j))
                    ;; j now length or next non-selected
                    (when (> j (1+ i))
                      (vector-push-extend #\: (the string %selstr))
                      (itos j %selstr))))
                (setf i j)
                finally
                (progn
                  (setf (aref (the string %selstr) 0) #\[)
                  (vector-push-extend #\] (the string %selstr)))))))
    %selstr))

(defun selections-to-reference (view &optional obj)
  ;; decompile selected listing entries into a reference
  (unless obj (setf obj (object-view-object view)))
  (let ((len (sequence-length view))
        (lb 0)
        (ub 0)
        (refs '()))
    (loop while (< lb len)
          do
          (setf ub (1+ lb))
          (if (cell-selected-p view 0 lb)
            (progn
              (loop while (and (< ub len)
                               (cell-selected-p view 0 ub))
                    do (incf ub))
              ;; ub is now at first non selected view or len
              (if (= ub (1+ lb))
                (push (makeref obj lb) refs)
                (push (makeref obj lb (if (= ub len) ':end (1- ub)))
                      refs))
              (setf lb (1+ ub)))
            (setf lb ub)))
    (nreverse refs)))

;;;
;;; mcl window operations
;;;

(defmethod ccl:select-all ((listing listing-view))
  (select-all-cells listing))

(defmethod ccl:clear ((listing listing-view))
  (deselect-all-cells listing))

(defmethod ccl:cut ((listing listing-view))
  (if (systemp (object-view-object listing))
    (warn "Can't cut from a system container.")
    (if (any-selection? listing)
      (listing-delete listing ':cut)
      (warn-user "No selections to cut!"))))

(defmethod ccl:copy ((listing listing-view))
  (remove-all-objects .pasteboard.)
  (add-objects (selected-objects listing)
               .pasteboard. nil :copy-first nil))

(defmethod ccl:paste ((listing listing-view))
  (let ((objects (loop for o in (container-objects .pasteboard.)
                       collect (duplicate-object o))))
    (when objects
      (if (> (sequence-nselected listing) 1)
        (warn-user "Can't paste objects to a region.")
        (listing-new listing 1 objects)))))

;;;
;;; listing-delete and listing-new are the edit functions
;;; for listing view
;;;

(defun listing-delete (listing command)
  ;; commands are :hide :unhide :delete, :undelete :expunge, :remove, :cut
  (let ((browser (view-browser listing))
        (object (object-view-object listing))
        (table (listing-view-displays listing))
        (cuts '())
        (cans '()))
    ;; keep these around until the dust settles...20.3.95
    (when (eq command ':remove) (error "shouldnt happen, :remove"))
    (unless (any-selection? listing) (error "shouldnt happen, nothing selected"))
    (map-selected-objects
     #'(lambda (o)
         (case command
           (:delete (delete-object o) (clear-display table o))
           (:undelete (undelete-object o) (clear-display table o))
           (:hide (hide-object o) (clear-display table o))
           (:unhide (unhide-object o) (clear-display table o))
           (:expunge (when (typep o 'id-mixin) (push o cans))
                     (delete-object o))
           (:cut (push (remove-object o object) cuts))))
     listing)
    (if (or (eq command ':expunge) (eq command ':cut))
      (progn
        (case command
          (:expunge (expunge-object object))
          (:cut
           (remove-all-objects .pasteboard.)
           (add-objects (nreverse cuts) .pasteboard.)))
        (set-table-sequence listing (container-objects object))
        (set-selection-typein browser "")
        (%deselect-all-cells listing)
        (when cans
          (unless (eq object .top-level.)
            ;; if containers were expunged outside top-level display
            ;; we have to update that as well.
            (let ((top (find-browser .top-level. 'listing-browser))
                  l)
              (when top
                (setf l (browser-listing top))
                (deselect-all-cells l)
                (set-table-sequence l (container-objects .top-level.))
                (update-browser top))))
          ;; close all browsers on deleted containers.
          (dolist (c cans)
            (map-capella-windows
             #'(lambda (b)
                 (unless (eq b browser)
                   (when (eq (object-view-object b) c)
                     (ccl:window-close b))))
             'browser)))
        (ccl:invalidate-view listing t))
      (draw-changed-displays listing))
    (update-browser browser)))

(defun listing-new (listing &optional (copies 1) type)
  ;; creates Copies number of new objects to be added 
  ;; after or before current selections.  If Type is a symbol
  ;; then new objects of that type are created and added at the 
  ;; position. If type is a list then it is the list of new
  ;; objects to add to the listing.  otherwise, if type is nil
  ;; the selected objects are duplcated after their current
  ;; positions. if nothing is selected new Type objects are appened
  ;; to the listing.
  (let ((browser (view-browser listing))
        (object (object-view-object listing))
        (append? (no-selection? listing))
        (updates '())
        (indirect? nil)
        (containers '())
        mode top?)
    (setf top? (eq object .top-level.))
    (if type 
      (if (consp type) 
        (setf mode ':paste copies (length type))
        (progn (when (functionp type) (setf indirect? t))
               (setf mode ':new)))
      (setf mode ':duplicate))
    (if append?
      (loop repeat copies
            for o = (if (eq mode ':new) 
                      (if indirect? (funcall type) (make-object type))
                      (pop type))
            collect o into olist
            unless (or top? (not (typep o 'id-mixin)))
            do (push o containers)
            finally (progn
                      (add-objects olist object)
                      (push (sequence-length listing) updates)))
      (map-selected-cells 
       #'(lambda (i)
           (let ((pos (if type i (1+ i)))
                 (new 
                  (loop repeat copies
                        for o = (case mode
                                  (:new (if indirect? (funcall type)
                                            (make-object type)))
                                  (:duplicate
                                   (duplicate-object
                                    (cell-contents listing 0 i)))
                                  (t (pop type)))
                        collect o
                        unless (or top? (not (typep o 'id-mixin)))
                        do (push o containers))))
             (%cell-deselect listing 0 i)
             (add-objects new object pos)
             (push pos updates)))
       listing))
    (set-table-sequence listing (container-objects object))
    (dolist (n updates)
      (loop for i from n repeat copies
            do (%cell-select listing 0 i)))
    (set-selection-typein browser listing)
    (ccl:invalidate-view listing t)
    (when containers (update-top-level :add containers))))

(defun listing-move (listing mode)
  ;; mode is :swap, :shuffle, :rotate, :reverse
  (let ((object (object-view-object listing))
        (lb (first-selection listing))
        (ub (last-selection listing)))
    (case mode
      (:reverse (retrograde-object object lb ub))
      (:swap (swap-object object lb ub))
      (:shuffle (shuffle-object object lb ub))
      (:rotate (rotate-object object lb ub)))
    ;; since moving the selections neither adds nor subtracts
    ;; nor deselects, we DONT want to call set-table-sequence,
    ;; we just want to update the actual sequence. 
    (setf (table-sequence listing) (container-objects object))
    (loop for i from (sequence-first-visible listing)
          to (sequence-last-visible listing)
          when (cell-selected-p listing 0 i)
          do (redraw-cell listing 0 i))))

;;;
;;; this is called by stella's update-top-level
;;;
(let ((ccl:*warn-if-redefine* nil)) ; stub in stella.lisp
  (defun update-top-level-display (browser mode objects)
    (let ((listing (browser-listing browser)))
      (case mode
        (:add
         (deselect-all-cells listing)
         (listing-new listing 1 objects))
        (:remove
         (deselect-all-cells listing)
         (set-table-sequence listing (container-objects .top-level.))
         (set-selection-typein browser "")
         (ccl:invalidate-view listing))
        (:change
         (let* ((table (listing-view-displays listing)))
           (dolist (o objects) (clear-display table o))
           (draw-changed-displays listing))))
      (update-browser browser)))
  )
;;;
;;; selection-view implements typein selection in listing views. typein
;;; buffer allows input to be either reference or expression. the
;;; "Select:" title is a button that deselects/selects all the objects.
;;;

(defclass selection-view (ccl:view)
  ()
  (:default-initargs 
    :view-nick-name ':selection-view
    :view-font (make-font :size 8 :defaults *capella-font*)))

(defclass refbuf (textbuf)
  ())

(defmethod initialize-instance :after ((view selection-view) &rest args)
  (declare (ignore args))
  (ccl:add-subviews
   view
   (make-instance 'button
     :view-position (xy 0 0)
     :view-size (xy 38 13)
     :title "Select:"
     :action #'clear/all
     :framed? nil 
     :default? nil
     :view-font *capella-font*
     :view-nick-name 'clear/all
     :background-color nil)
   (make-instance 'listing-status-view :view-position (xy 0 0)
                  :view-size (xy 125 12))
   (make-instance 'refbuf
     :view-position (xy 5 15) 
     :view-size (xy 0 12)
     :title "" ; width is computed
     :action nil
     :view-font *capella-font*
     :allow-returns t 
     :framed? t
     :view-nick-name ':typein)))

(defclass listing-status-view (ccl:view)
  ((status :initform (cons 0 0) :accessor listing-status-view-status))
  (:default-initargs
    :view-font (make-font :size 8 :defaults *capella-font*) 
    :view-nick-name ':status))

(defun update-listing-status (browser)
  (let* ((listing (browser-listing browser))
         (view (ccl:view-named ':status 
                               (ccl:view-named ':selection-view browser)))
         (data (listing-status-view-status view)))
    (setf (car data) (sequence-length listing)
          (cdr data) (sequence-nselected listing))
    ;(ccl:invalidate-view view nil)
    (ccl:view-draw-contents view)))

(defmethod ccl:view-draw-contents ((view listing-status-view))
  (let ((status (slot-value view 'status)))
    (ccl:with-font-focused-view view
      
      (ccl:rlet ((r :rect :topleft (xy 0 0) :bottomright 
                    (ccl:view-size view)))
        (ccl:with-fore-color *bgc*
          (#_fillrect r ccl:*black-pattern*)))
      (#_moveto 2 9)
      (ccl:with-pstrs ((s1 (fi2s (car status) %posbuf 4))
                       (s2 (if (= (car status) 1) " object, " " objects, ")))
        (#_DrawString s1)
        (#_DrawString s2))
      (#_moveto 60 9)
      (ccl:with-pstrs ((s1 (fi2s (cdr status) %posbuf 4))
                       (s2 " selected."))
        (#_DrawString s1)
        (#_DrawString s2)))))

(defun listing-selection-view (browser)
  (ccl:view-named ':selection-view browser))

(defmethod ccl:set-view-size :after ((view selection-view) x &optional y)
  (unless y (setf y (xy-y x) x (xy-x x)))
  (let ((typein (ccl:view-named ':typein view))
        (status (ccl:view-named ':status view)))
    (ccl:set-view-position status
                           (xy (- (view-width view) 
                                  (view-width status) )
                               0))
    (ccl:set-view-size typein (- x 8) (view-height typein)))) 

(defmethod ccl:view-key-event-handler ((view refbuf) char)
  ;; return set arg.
  (if (or (char= char #\Newline) (char= char #\Enter)) 
    (selection-typein-action view)
    (call-next-method)))

(defun selection-typein-action (typein)
  ;; activated by #\Newline in the selection typein buffer.
  ;; deselect whatever is currently selected and, if
  ;; buffer has anything, select it.
  (let ((s (textbuf-text typein))
        (l (browser-listing (view-browser typein))))
    (unless (string= s "")
      (if (char= (elt s 0) #\()
        (if (balanced-form? s)
          (ccl:with-cursor ccl:*watch-cursor*
            (deselect-all-cells l)
            (setf (listing-view-reference l) nil)
            (select-by-expr s l))
          (warn-user "\"~A\" is a bad selection." s))
        (let* ((r (parse-references s l)))
           (if r
             (ccl:with-cursor ccl:*watch-cursor*
               (deselect-all-cells l)
               (mappos #'(lambda (i l) (%cell-select l 0 i)) r)
               (setf (listing-view-reference l) r))
             (warn-user "\"~A\" is a bad selection." s))))
      (update-browser (view-browser typein)))))

(defun select-by-expr (input listing)
  (let ((cell 0)
        (view listing)
        (func    ; silly -- why do i have to compile in mcl???
         (coerce
          `(lambda ($)
             (declare (special view cell))
             (catch :no-value
               (when ,(walk-map-expr
                       (read-from-string input) nil nil nil)
                 (%cell-select view 0 cell)))
             (incf cell))
          'function)))
    (declare (special cell view))
    (map-sequence-objects func view)))
        
(defun clear/all (button)
  ;; clear any selected, otherwise select all.
  (let* ((b (view-browser button))
         (l (browser-listing b))
         (s ""))
    (if (any-selection? l)
      (deselect-all-cells l)
      (progn (select-all-cells l)
             (setf s "*")))
    (set-selection-typein b s)
    (update-browser b)))

(defun set-selection-typein (browser source)
  ;; source is either string or listing view. in either case, we put
  ;; the selections in the typeing buffer and null out any user specified
  ;; reference.
  (if (stringp source)
    (setf (listing-view-reference (browser-listing browser)) nil)
    (progn
      (setf (listing-view-reference source) nil)
      (setf source (selections-to-string source))))
  (let ((buf (ccl:view-named ':typein
                             (ccl:view-named ':selection-view browser))))
    (let ((b (ccl:fred-buffer buf)))
      (when (> (ccl:buffer-size b) 0)
        (ccl:buffer-delete b 0 (ccl:buffer-size b)) )
      (ccl:buffer-insert b source 0)
      (ccl:fred-update buf))))
    
;;;
;;;
;;;

(defclass listing-control-view (gray-3d-framed-view)
  ()
  (:default-initargs
    :view-nick-name ':control-view))

(let ((tools 
       (list 
        (make-tool :name ':open :icon-id #x4000
                   :action #'(lambda (b) b (open-browser)))
        (make-tool :name ':info :icon-id #x4008
                   :action #'(lambda (b) b (open-browser 'info-browser)))
        (make-tool :name ':code :icon-id #x4001
                   :action #'(lambda (b) b (open-browser 'algorithm-browser)))
        (make-tool :name ':set :icon-id #x4002
                   :action #'(lambda (b) b (worksheet 'set-worksheet)))
        (make-tool :name ':struc :icon-id #x4003
                   :action #'(lambda (b) b (worksheet 'move-worksheet)))
        (make-tool :name ':new :icon-id #x4004
                   :action #'(lambda (b) b (worksheet 'new-worksheet)))
        (make-tool :name ':out :icon-id #x4005
                   :action #'(lambda (b) b (worksheet 'out-worksheet)))
        ;(make-tool :name ':mix  :icon-id #x4006
        ;           :action #'(lambda (b) b (worksheet 'mix-worksheet)))
        ;(make-tool :name ':seq :icon-id #x4007
        ;           :action #'(lambda (b) b (worksheet 'sequence-worksheet)))
        (make-tool :name ':layout :icon-id #x400c
                   :action #'(lambda (b) b (worksheet 'output-worksheet)))
        (make-tool :name ':query  :icon-id #x4009
                   :action #'(lambda (b) b (worksheet 'query-worksheet)))
        (make-tool :name ':load :icon-id #x400a
                   :action #'(lambda (b) b (menubar-load)))
        (make-tool :name ':save :icon-id #x400b
                   :action #'(lambda (b) b (menubar-archive)))))
      view)

(defun make-listing-control-view ( &rest args)
    (setf view
          (apply #'make-tool-view
                 tools :view-class 'listing-control-view
                 :tool-size 18 :spacing 4 :offset (xy 4 4) args))
    view))

(defparameter %check-selections-limit% 25)

(defun update-listing-controls (browser)
  (let* ((controls (ccl:view-named ':control-view browser))
         (object (object-view-object browser))
         (listing (browser-listing browser))
         (selected (sequence-nselected listing))
         (toplevel? (eq object .top-level.))
         (system? (systemp object))
         (selclass (and (> selected 0)
                        (selection-class? listing %check-selections-limit%)))
         (container? (if (= selected 0)
                       (class-name (class-of object))
                       (and selclass 
                            (find (class-name selclass)
                                  '(thread merge container
                                     generator )))))
         (algorithm? (and (= selected 1)
                          selclass
                          (find (class-name selclass)
                                '(generator algorithm )))))
    ;; Open icon
    (if (and (= selected 1) container?)
      (control-enable (ccl:view-named ':open controls))
      (control-disable (ccl:view-named ':open controls)))
    
    ;; Info icon
    (if (< selected 2)
      (control-enable (ccl:view-named ':info controls))
      (control-disable (ccl:view-named ':info controls)))

    ;; Code icon
    (if algorithm?
      (control-enable (ccl:view-named ':code controls))
      (control-disable (ccl:view-named ':code controls)))
    
    ;; Edit icon
    (if (or (not system?) (and (> selected 0) container?))
      (control-enable (ccl:view-named ':set controls))
      (control-disable (ccl:view-named ':set controls)))

   ; (if (or (not toplevel?) (not (eq algorithm? 'algorithm)))
   ;   (control-enable (ccl:view-named ':set controls))
   ;   (control-disable (ccl:view-named ':set controls)))
   ; 
    ;; New icon
    (if (or toplevel? (not system?))
      (control-enable (ccl:view-named ':new controls))
      (control-disable (ccl:view-named ':new controls)))
    
    ;; Out icon
    (if (> selected 0)
      (control-enable (ccl:view-named ':out controls))
      (control-disable (ccl:view-named ':out controls)))
#|    
    ;; Mix and Save icons
    (if (or (and toplevel? (> selected 0))
            (and (not system?) container?))
      (progn
        (control-enable (ccl:view-named ':mix controls))
        (control-enable (ccl:view-named ':save controls)))
      (progn
        (control-disable (ccl:view-named ':mix controls))
        (control-disable (ccl:view-named ':save controls))))
    
    ;; Seq icon
    (if (if toplevel? (> selected 0) (not system?))
      (control-enable (ccl:view-named ':seq controls))
      (control-disable (ccl:view-named ':seq controls)))
|#
    
    ;; Layout icon
    (if (or (and toplevel? (> selected 0))
            (and (not system?) container?))
      (control-enable (ccl:view-named ':layout controls))
      (control-disable (ccl:view-named ':layout controls)))
    
    ;; Query icon
    (if (or (and toplevel? (> selected 0))
            (not system?))
      (control-enable (ccl:view-named ':query controls))
      (control-disable (ccl:view-named ':query controls)))
    
    ;; Load icon
    (if (or toplevel? (not system?))
      (control-enable (ccl:view-named ':load controls))
      (control-disable (ccl:view-named ':load controls)))))
  
;(defun open-ok? (&optional (browser *browser*))
;  (and (listing-browser? browser)  (any-selection? browser)))

(defun open-ok? (&optional (browser *browser*)) 
  (if browser
    (let ((object (object-view-object browser)))
      (if (listing-browser? browser)
        (let ((l (browser-listing browser)))
          (if (any-selection? l)
            (eq (selection-class? l) (find-class 'container))
            (typep object 'container)))
        (typep object 'container)))
    nil))

(defun set-ok? (&optional (browser *browser*))
  ;; if top level something must be selected.
  (if (listing-browser? browser)
    (if (eq (object-view-object browser) .top-level.)
      (any-selection? browser)
      t)
    nil))

(defun new-ok? (&optional (browser *browser*))
  (listing-browser? browser))

#|
(defun out-ok? (&optional (browser *browser*))
  (and (listing-browser? browser) (any-selection? browser)))

(defun mix-ok? (&optional (browser *browser*))
  (if browser
    (let ((object (object-view-object browser)))
      (if (eq object .top-level.)
        (any-selection? browser)
        (if (systemp object)
          nil
          (if (listing-browser? browser)
            (let ((l (browser-listing browser)))
              (if (any-selection? l)
                (eq (selection-class? l %check-selections-limit%) 
                    (find-class 'id-mixin)) 
                (typep object 'id-mixin)))
            (typep object 'id-mixin)))))
    nil))

(defun seq-ok? (&optional (browser *browser*))
  (if browser
    (let ((object (object-view-object browser)))
      (if (eq object .top-level.)
        (any-selection? browser)
        (not (systemp object))))
    nil))
|#

(defun info-ok? (&optional (browser *browser*)) browser)

(defun code-ok? (&optional (browser *browser*)) 
  (if browser
    (let ((object (object-view-object browser)))
      (if (listing-browser? browser)
        (let ((l (browser-listing browser)))
          (if (any-selection? l)
            (eq (selection-class? l) (find-class 'algorithm))
            (typep object 'algorithm)))
        (typep object 'algorithm)))
    nil))

(defun query-ok? (&optional (browser *browser*))  
  (if (listing-browser? browser)
    (let ((object (object-view-object browser)))
      (if (eq object .top-level.)
        (any-selection? browser)
        (not (systemp object))))
    nil))

(defun archive-ok? (&optional (browser *browser*))  
  (if browser
    (let ((object (object-view-object browser)))
      (if (eq object .top-level.)
        (any-selection? browser)
        (if (systemp object)
          nil
          (if (listing-browser? browser)
            (let ((l (browser-listing browser)))
              (if (any-selection? l)
                (eq (selection-class? l %check-selections-limit%) 
                    (find-class 'id-mixin)) 
                (typep object 'id-mixin)))
            (typep object 'id-mixin)))))
    nil))

(defun load-ok? (&optional (browser *browser*)) browser t)



