//
//
// File:         pmcl-midi.cc
//
// Summary:      Implementation file for the pmclmidi shared-library 
//               reimplementation of the Apple MidiManager interface for the
//               PPC version of Macintosh Common Lisp (3.9 or higher).
//
// Dependencies: pmcl-midi.hh
//               pmcl-midi.lisp
//               pmcl-entry.lisp
//
//               ***********************************************************
//               Copyright (c) 1996 by Tobias Kunze and Rick Taube.  All
//               rights reserved.  Use and copying of this software and
//               preparation of derivative works based upon this software
//               are permitted and may be copied as long as no fees or
//               compensation are charged for use, copying, or accessing
//               this software and all copies of this software include this
//               copyright notice.  Suggestions, comments and bug reports
//               are welcome.  Please address email regarding this midi
//               implementation to: t@kunze.stanford.edu or
//               hkt@ccrma.stanford.edu.
//               ***********************************************************
//
// Author:       Tobias Kunze
// E-Mail:       t@kunze.stanford.edu
// Org:          CCRMA, Stanford University
//
// Orig-Date:    15-May-96 at 17:49:41
// Last-Mod:     26-Jul-96 at 00:15:35 by Tobias Kunze
//
// Description:  See pmcl-midi.hh
//
//               
//#ident "$Revision: 1.0 $"
//
// Changes:
//    15-May-96  tk    rewrite as a shared PPC library for PPC-MCL (3.9) 
//    07-Jul-96  tk   Revision 1.0


#include "pmcl-midi.hh"
#include <assert.h>

const long kChannelTag = 0x04000000L; // bits 26-27 == '01'

void main() {}                        // make the CW 8.0 linker happy

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// Dynamic Invisible Ports Management
//
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

//
// Invisible (input) port's refnums are stored in the array fInvInPorts.
// A refnum of 0 (kNoPort) means the port does not exist. We allocate new
// InvInPorts on overflow conditions and deallocate them after they have been
// idling around for fBufIdleTimeOut.
// I am not sure I'll support dynamic allocation/deallocation in the future. 
// Static allocation of queues grabs only 650K max, anyway.


//
// Initialize
// ----------
// Called during Construction and C (Re-)Initialization (MidiDrSetup)

void DynamicPorts::Init(long bufidletimeout, long wakeupperiod) {
  fBufIdleTimeOut = bufidletimeout;
  fWakeUpPeriod = wakeupperiod;
  for (int i = 0; i < kMaxInvInPorts; i++) {
    fInvInPorts[i] = kNoPort;
    fObsInPorts[i] = kNoPort;
  }
  fInvInPorts[kMaxInvInPorts] = kNoPort; // initialize last stop field, too
}


//
// Construction
// ------------

DynamicPorts::DynamicPorts()
  : bOutputInProgress(FALSE), fNumInvInPorts(0), fFirstFreeInObsInPorts(0) 
{
  Init(5000, 10000); 
}


//
// FlushInvInPorts
// ---------------

void DynamicPorts::FlushInvInPorts() {
  for (int i = 0; i <= fNumInvInPorts; i++) {
  	if (fInvInPorts[i] == 0)
  	  return;
  	else 
      MIDISetOffsetTime((short)fInvInPorts[i], midiGetEverything);
  }
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// Midi Driver Configuration and Interface Implementation
//
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------


//
// Constructor
// -----------

MidiDrConfig::MidiDrConfig(DynamicPorts* ports) 
  : bFlushing(FALSE), pDynPorts(ports),               // , pLispReadHook(0x0L) 
    fTimePort(0), fInPortA(0), fInPortB(0), fOutPortA(0), fOutPortB(0),
    fInvOutPort(0) {     
}



// ---------------------------------------------------------------------------
// 
// Private Functions
//


//
// IsChannelMessage
// ----------------
// My little home-brew test.

inline int MidiDrConfig::IsChannelMessage(MIDIPacketPtr packet) {
  return (packet->data[0] >= 0x80 && packet->data[0] < 0xF0);
}


//
// WriteEncodedPacket
// ------------------
// Output Midi messages: convert channel messages with port encoding to
// standard midi messages and output them on their respective port.  Output
// all other mesages on BOTH ports!
// NOTE: There is currently NO error checking for the latter case, i.e.,
// packets will be lost on overflow conditions.

inline int MidiDrConfig::WriteEncodedPacket(MIDIPacketPtr packet) {
  short port;
  if (IsChannelMessage(packet)) { // don't check other message types!
    if (packet->data[1] & 0x80) { // is bit 15 (the `port' bit) set?
      port = fOutPortB;
    } else {
      port = fOutPortA;
    }
    return MIDIWritePacket(port, packet);
  } else {
    MIDIWritePacket(fOutPortA, packet);
    MIDIWritePacket(fOutPortB, packet);
    return kMidiSuccess;
  }
}




// ---------------------------------------------------------------------------
//
// The Instances
//

DynamicPorts Queues = DynamicPorts();
MidiDrConfig Driver = MidiDrConfig(&Queues);




#pragma export on

// --------------------------------------------------------------------------
// 
// DynamicPorts LISP Entry Points
//


//
// RegisterInvInPort
// -----------------
// Add a new dynamic input queue.  This is a client-only function and may not
// be called at interrupt time.  Returns the newly added refnum or -1 if the
// array is full.

int RegisterInvInPort(short refnum) {
  int inports = Queues.fNumInvInPorts;

  if (inports >= kMaxInvInPorts)
    return -1;
  else {
    if (inports > 0) {
      int i,j;	
  	  // make room by shifting nums back
  	  for (i = inports - 1, j = inports; i >= 0; i--, j--)
        Queues.fInvInPorts[j] = Queues.fInvInPorts[i];
    }
    Queues.fInvInPorts[0] = refnum;
    Queues.fNumInvInPorts++;
    return refnum;
  }
}


//
// RotateInvInPorts
// ----------------
// Rotate the vector by idx places, i.e. 1 rotates it "left".
// It's probably not worth the hassle, but this algorithm shifts in place
// and is robust enough to take any integer as input.

void RotateInvInPorts(int idx) {
  int inports = Queues.fNumInvInPorts;
  
  if (idx >= inports)           // make sure idx is within bound (modulo)
    idx = idx % inports;
  if (idx < 0)                  // and non-negative or we segmentation-fault
    idx += inports;

  int ito, ifrom, istored;
  int vtmp;
  int count = 0;                // count copy operations

  for (ito = 0, ifrom = idx; ; ) {
    vtmp = Queues.fInvInPorts[ito];    // store away the startelement
    istored = ito;              // and remember its index

    // now loop while copying the item either from ifrom of from vtmp if we
    // hit the stored item (which closes a subcycle in the loop)
    for ( ; ; ) {
      if (ifrom == istored) {
        Queues.fInvInPorts[ito] = vtmp;
        count++;
        // if we copied from vtmp, increment ifrom and ito to start a new
        // subcycle
        ifrom = (ifrom + 1) % inports;
        ito = (ito + 1) % inports;
        break;
      } else {
        Queues.fInvInPorts[ito] = Queues.fInvInPorts[ifrom];
        count++;
        ito = ifrom;
        ifrom = (ifrom + idx) % inports;
      }
    }
    if (count >= inports)
      break;
  }
}


//
// RemoveObsPorts
// --------------
// This function is called from nearly all %mm-functions.
// I know this may lead to dangling empty buffers if the user plays a lot of
// events and then leaves the port open without doing any midi afterwards,
// but there is nothing I can do (except hacking the MCL event mechanism). 

void RemoveObsPorts() {
  if (Queues.fFirstFreeInObsInPorts > 0) {
    int port;
    for (int i = 0; i < kMaxInvInPorts; i++) {
      port = Queues.fObsInPorts[i];
      if (port > 0)
        MIDIRemovePort(port);
      else
        break;                    // all ports are at the beginning
    }
    Queues.fFirstFreeInObsInPorts = 0;
  }
}


//
// GetbOutputInProgress
// --------------------
// Return the address of bOutputInProgress to LISP.

unsigned char* GetbOutputInProgress() {
  return &(Queues.bOutputInProgress);
}



// ------------------------------------------------------------------------
// 
// DynamicPorts Toolbox Callback Functions
//


// MarkObsInPorts
// --------------
// Once CommonMusic has signed into the MidiManager, it schedules a periodic
// callback with a time interval of fWakeUpPeriod.  The callback checks
// whether a buffer should be considered expired by looping through the
// InvInPorts looking at their RefCon field (which holds the time stamp of
// the last message written to that port).  If a time stamp is more than
// fBufIdleTimeout ms old, it's InvInPort's ID gets moved to fObsInPorts for
// future deallocation from Lisp (we can't deallocate at interrupt time).
// We don't remove any ports, though, while midi output is in progress (while
// the bOutputInProgress flag is TRUE.  In this case, we simply wait until
// we're called again. 
// We schedule the WakeUp with our invisible output port (it doesn't matter,
// anyways) and simply ignore the myRefCon field.
// Also: we NEVER remove the current invisible input port, whose refnum is
// given by convention by the first element of fInvInPorts[]. 

extern pascal void MarkObsInPorts(long curTime, long myRefCon) {
  long dummy = myRefCon;
  int i, m, n, port;
  int numinvports = Queues.fNumInvInPorts;
  long ExpirationTime = curTime - Queues.fBufIdleTimeOut;
  if (numinvports > 1) {         // don't bother deallocating the last InvInPort
    if (Queues.bOutputInProgress) {
      return;
    } else {
      for (i = 1; i < numinvports; ) { // don't fuck with fInvInPorts[0];
        port = Queues.fInvInPorts[i];
        // remove if expired; again, don't bother if it's the last buffer
        if (numinvports > 1 && MIDIGetRefCon(port) <= ExpirationTime) {
          // append to fObsInPorts[]
          // assert(Queues.fFirstFreeInObsInPorts < kMaxInvInPorts);
          Queues.fObsInPorts[Queues.fFirstFreeInObsInPorts] = port;
          Queues.fFirstFreeInObsInPorts++;
          // close gap in fInvInPorts[]
          for (m = i, n = 1 + i; n < numinvports; m++, n++)
            Queues.fInvInPorts[m] = Queues.fInvInPorts[n];
          // clear the last element (re-decrement first)
          Queues.fInvInPorts[n-1] = kNoPort; 
          numinvports--;
        } else
          i++;                  // go check next InvInPort
      }
      Queues.fNumInvInPorts = numinvports; // update the class field
    }
  }
}





// --------------------------------------------------------------------------
// 
// MidiDrConfig LISP Entry Points
//

// 
// MidiReadEvent
// -------------
// The entry point for read requests from Lisp.  It translates into a call to
// MIDIPoll with the mode set to midiGetCurrent (instead of the standard
// midiGetNothing).
// If MidiPoll wouldn't be a &^#@!% Pascal PROCEDURE and return at least an
// OSErr, I could test for empty buffers here...

int MidiReadEvent() {
  MIDIPoll(Driver.fInPortA,midiGetCurrent);
  MIDIPoll(Driver.fInPortB,midiGetCurrent);
  return kMidiSuccess;
}


// 
// SetReadHook
// -----------
// This is called every time a invisible input port is created. 

void SetReadHook(short refNum) {
  MIDIReadHookUPP InvReadHookPtr = NewMIDIReadHookProc(&InvReadHook);
  MIDISetReadHook(refNum,(ProcPtr)InvReadHookPtr);
}


//
// FlushQueues
// -----------
// Set bFlushing to TRUE and schedule a one-time callback to switch it back
// off after fFlushDuration ms.
// Since the Midi Manager provides no way to determine whether an input
// port's buffer is empty, this approach may lead to events still dripping
// out of the port after flushing mode ended if there were _huge_
// amounts of future packets awaiting delivery that could not be flushed
// within the given time.

void FlushQueues() {
  Driver.bFlushing = TRUE;
  Queues.FlushInvInPorts();
  MIDITimeUPP EndFlushingPtr = NewMIDITimeProc(&EndFlushing);
  MIDIWakeUp(Driver.fTimePort,
             (long)(MIDIGetCurTime(Driver.fTimePort) + Driver.fFlushDuration), 
             0L, (MIDITimeUPP)EndFlushingPtr);
}
 

//
// MidiDrSetUp
// -----------
// The main initialization routine.

short* MidiCSetUp(short timeport, short inportA, short inportB, short outportA,
                  short outportB, short invoutport,
                  void (*lispreadhook) (long, long),
                  long wakeupperiod, long bufidletimeout, long flushduration) 
{

  Queues.Init(bufidletimeout, wakeupperiod);

  // Driver Initialization 
  Driver.fTimePort = timeport;
  Driver.fInPortA = inportA;
  Driver.fInPortB = inportB;
  Driver.fOutPortA = outportA;
  Driver.fOutPortB = outportB;
  Driver.fInvOutPort = invoutport;
  Driver.fFlushDuration = flushduration;
  Driver.pLispReadHook = lispreadhook;

  // Set Readhooks and RefCon fields
  MIDIReadHookUPP ReadHookPtr = NewMIDIReadHookProc(&ReadHook);
  MIDISetReadHook(inportA, (ProcPtr)ReadHookPtr);
  MIDISetRefCon(inportA, (long)inportA);
  MIDISetReadHook(inportB, (ProcPtr)ReadHookPtr);
  MIDISetRefCon(inportB, (long)inportB);
  
  // Schedule a callback to release empty buffers 
  MIDITimeUPP MarkObsInPortsPtr = NewMIDITimeProc(&MarkObsInPorts);
  MIDIWakeUp(timeport,0,wakeupperiod,(MIDITimeUPP)MarkObsInPortsPtr);
  // or simply:
  // MIDIWakeUp(timeport,0,wakeupperiod,NewMIDITimeProc(&MarkObsInPorts));
  
  return &(Queues.fInvInPorts[0]);
}




// ------------------------------------------------------------------------
// 
// MidiDrConfig Toolbox Callback Functions
//


//
// ReadHook
// --------------
// This read hook gets called by the MidiManager for each incoming message.
// Since LISP can't be called from within an interrupt, I set both input
// ports to respond only to calls to midiPoll(input_port,midiGetCurrent),
// which are issued from LISP at non-interrupt time.  This is done by setting
// the input port's offset time to midiGetNothing.
// Since lots of MidiManager error packets are waiting after our various
// output buffer overflows, return midiMorePacket to quickly flush all 
// unknown message types. Also, according to the Common Music documentation 
// all pending messages are read in response to one single midireadevent call
// (see cm/doc/midi.html).  Thus, return midiMorePacket for known messages,
// too.
// refCon has to contain the port refnum.

// *** I don't understand how the MidiManger passes the first (the
// *** message-type) byte.  It seems like it throws away the first byte...
// *** How do I find out what a channel message is and what not?  Looks like
// *** I have to test the upper nibble of the status byte to be between 1000 
// *** and 1110.

extern pascal short ReadHook(MIDIPacketPtr packet, long refCon) {
  long message, time;

  // assert(refCon == Driver.fInPortA || refCon == Driver.fInPortB);
  if (// not a continued message
      ((packet->flags & midiContMask)==midiNoCont) &&
      // packet contains Midi data
      ((packet->flags & midiTypeMask)==midiMsgType) &&
      // use time stamp in packet
      ((packet->flags & midiTimeStampMask)==midiTimeStampValid)) {
    message = (//((kChannelTag << 26) & 0xc000000) |
               kChannelTag |
               ((packet->len-6 << 24) & 0x03000000) |
               ((packet->data[0] << 16) & 0xff0000) |
               ((packet->data[1] <<  8) & 0xff00) |
               (packet->data[2] & 0xff));
    // mark channel messages coming from port b as such
    if (refCon == Driver.fInPortB && Driver.IsChannelMessage(packet))
      message = message | 0x8000; // set the port bit (bit 15)
    time=packet->tStamp;
    (*(Driver.pLispReadHook)) (message, time); 
    return midiMorePacket;
  }
  return midiMorePacket;
}


//
// InvReadHook
// -----------------
// The read hook for invisible ports. This hook is called whenever future
// midi output stored in the invisible input buffer becomes current. The hook
// simply passes the current message to the real output port, decoding the
// port bit, if necessary.
// During output queue flushing, however, the hook checks the message type
// and passes only note-off's and note-on's with zero velocity, ignoring all
// other message types. Thus, multiple notes hanging on the same channel and
// key will be turned off properly.  It does NOT pass on Sysex and other
// multi-packet messages.  Don't know whether this is a bug.

extern pascal short InvReadHook(MIDIPacketPtr packet, long myRefCon) {
#pragma unused (myRefCon)
  // when flushing, block all messages except for note-off's and note-on's
  // with zero velocity and OR the midiTimeStampCurrent bit in the flags
  // field to make the message current
  if (Driver.bFlushing) {
    if (! (Driver.IsChannelMessage(packet) &&
           ((packet->data[0] & 0xf0)  == 0x80 ||
            (packet->data[0] & 0xf0) == 0x90 && (packet->data[2] == 0))))
        return midiMorePacket;  // ignore message
    packet->flags |= midiTimeStampCurrent;
  }
  int err = (Driver.WriteEncodedPacket(packet));
  if (err == 0) {
    return midiMorePacket;
  }
  return midiKeepPacket;
}


//
// EndFlushing
// -----------
// Terminates flushing mode after fFlushDuration milliseconds.  This callback
// takes a terminator-style approach to end flushing mode by calling
// MIDIFlush on all invisible input ports, which may lead to hanging notes
// under rare circumstances.

extern pascal void EndFlushing(long curTime, long RefCon) {
#pragma unused (curTime, RefCon)
  short port;
  int i;
  
  for (i = 0; i < kMaxInvInPorts ; i++) {
    port = Driver.pDynPorts->fInvInPorts[i];
    if (port == 0)
      break;
    MIDIFlush(port);
    MIDISetOffsetTime(port, midiGetCurrent);
  }
  Driver.bFlushing = FALSE;
}

#pragma export off

//
// -*- EOF -*-
