;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :cm)

(defconstant +message-sizes+ (vector 3 3 3 3 2 2 3 0))
(defconstant +midifile-header-length+ 14)
(defconstant +miditrack-header-length+ 8)
(defconstant +MThd+ 1297377380)  ; 'MThd'
(defconstant +MTrk+ 1297379947)  ; 'MTrk'
(defconstant +eot+  ':eot)       ; end of track signal

(defstruct mfinfo                      ; midi file parsing info
  (scaler 1)                           ; scaler * delta = time
  (status 0)                           ; last status byte read
  (ttype :seconds)                     ; :ticks, :milliseconds, :seconds
  (size 0)                             ; size of last message parsed 
  (tracks 1)                           ; number of tracks 
  (track -1)                           ; current track 
  (tracklen 0)                         ; current track's length 
  (divisions 96)                       ; ticks per midi quarter 
  (format 0)                           ; file format size 
  (chanmap nil))                       ; optional vector channel map

(defstruct (midifile (:include mfinfo))
  pathname
  stream
  (direction ':input))

(defstruct (midimix (:include midifile))
  (output-time 0.0)    ; (+ (- input-time start) output-start)
  (output-start 0.0)   ; output time offset
  (input-time 0.0)     ; file time
  (start 0.0)          ; start time in input file to begin mixing
  end                  ; last note on time in input to process
  message
  (channels nil)
  predicate
  filter
  hash)

(defun make-queue () (list nil))
(defmacro queue-head (q) `(car ,q))
(defmacro queue-last (q) `(cdr ,q))

(defun set-scaler (info usecs &aux (format (mfinfo-ttype info)) )
  (setf (mfinfo-scaler info)
        (cond ((or (null format) (eq format ':ticks)) 1)
              ((eq format ':seconds) 
               (/ usecs (* (mfinfo-divisions info) 1000000.0))) 
              ((eq format ':milliseconds)
               (/ usecs (* (mfinfo-divisions info) 1000.0))
               ;(round usecs (* (mfinfo-divisions info) 1000.0))
               ) 
              (t (error "~S is not a time format." format)))))

(defun set-ttype (info ttype)
  (setf (mfinfo-ttype info) ttype)
  (set-scaler info 5000000)
  ttype)

(defun midifile-set-time-format (midifile format)
  (set-ttype midifile format))

;;;
;;; midifile opening and closing
;;;

(defun midifile-open (file &rest args &key direction (track -1 tr?)
                                          (ttype ':ticks tt?)
                                          chanmap)
  (cond ((typep file 'midifile)
         (when direction (setf (midifile-direction file) direction))
         (when tr? (setf (midifile-track file) track))
         (when tt? (setf (midifile-ttype file) ttype)))
        (t
         (let ((f (merge-pathnames file *default-midi-pathname*)))
           (when (probe-file f) (setf file f))
           (setf file (apply #'make-midifile :pathname file args)))))
  (when (consp chanmap)
    (setf (mfinfo-chanmap file) (apply #'vector (mfinfo-chanmap file))))
  (midifile-open-stream file)
  (when (> (midifile-track file) -1)
    (midifile-open-track file (midifile-track file) (midifile-ttype file)))
  file)

(defun midifile-close (midifile)
  (if (streamp (midifile-stream midifile))
    (progn (when (midifile-track-open? midifile)
             (midifile-close-track midifile))  
           (close (midifile-stream midifile))
           (setf (midifile-stream midifile) nil)
           t)
    nil))

(defmacro with-open-midifile ((f path &rest args) &body forms)
  `(let (,f)
     (unwind-protect 
       (progn
         (setf ,f (midifile-open , path ,@ args))
         ,@ forms)
      (midifile-close ,f))))

(defun midifile-open-stream (midifile &aux (info midifile))
  ; open file, read/write header information
  (setf (midifile-stream midifile)
        (open (midifile-pathname midifile)
              :element-type '(unsigned-byte 8)
              :direction (midifile-direction midifile)
              :if-exists ':supersede 
              :if-does-not-exist
              (if (eq (midifile-direction midifile) ':output)
                  ':create ':error)))
  (if (eq (midifile-direction midifile) ':input)
      (let (format tracks divisions)
        (multiple-value-setq (format tracks divisions)
          (%midifile-read-header (midifile-stream midifile)))
        (setf (mfinfo-format info) format)
        (setf (mfinfo-tracks info) tracks)
        (setf (mfinfo-divisions info) divisions)
        (setf (mfinfo-scaler info) 1))             ; assume :ticks
      (%midifile-write-header (midifile-stream midifile)
                              0 1 (mfinfo-divisions info)))
  ;(setf (midifile-track midifile) -1)
  (midifile-stream midifile))

;;;
;;; midifile track opening and closing
;;;

(defun midifile-track-open? (midifile)
  (/= (mfinfo-track midifile) -1))   ; this wont work 

(defun midifile-open-track (midifile track &optional (ttype :ticks))
  ;; move to track, parse track header and position file pointer
  ;; to start of data.  file must be open and track must be a valid.
  (let ((stream (midifile-stream midifile)))
    (midifile-set-time-format midifile ttype)
    (setf (mfinfo-track midifile) track)
    (cond ((eq (midifile-direction midifile) ':input)
           (%midifile-set-track stream track)
           (setf (mfinfo-tracklen midifile)
                 (%midifile-read-track-header stream))
           )
          (t
           (unless (and (= track 0) (= (midifile-tracks midifile) 1))
             (error "Multi track output files not supported."))
           (setf (mfinfo-tracklen midifile) 0)
           (setf (mfinfo-track midifile) 0)
           (file-position stream +midifile-header-length+)
           (%midifile-write-track-header stream 0)))
    stream))

(defun midifile-close-track (midifile)
  (when (eq (midifile-direction midifile) ':output)
    (let ((stream (midifile-stream midifile)))
      (when (= (file-position stream) (file-length stream))
        (%midifile-write-message (make-eot) stream 0))
      (%midifile-set-track stream 0)
      (%midifile-write-track-header
        stream
        (- (file-length stream)
           (+ +midifile-header-length+ +miditrack-header-length+)))))
  (setf (mfinfo-track midifile) -1))

(defun %midifile-set-track (stream track)
  ;; set file position to tbe begining of track header
  (loop for p = +midifile-header-length+ 
        then (+ p 8 (%midifile-read-track-header stream)) ; 8='mthd'+lenBytes
        for c from 0
        do (file-position stream p)
        while (< c track))
  track)

;;;
;;; midifile track processing
;;;

; :ma (dotrack i f  :ticks nil nil)
; :ma (dotrack i f   :seconds nil nil)
; :ma (dotrack i f   :milliseconds beg end)
; :ma (dotrack i nil :seconds nil start)

(defmacro dotrack (info function ttype start end)
  ;; return optimized loop based on the _compile time_ value of ttype
  ;; (time type) and if start and end varibles are present.
  (let ((fil (gensym))
        (inf (gensym))
        (q (gensym)) (m (gensym)) (s (gensym)) (x (gensym)) 
        d f scl)
    (ecase ttype 
      (:seconds (setf d 'float scl `(mfinfo-scaler ,inf)))
      (:ticks (setf d 'fixnum scl nil))
      (:milliseconds (setf d 'fixnum scl `(mfinfo-scaler ,inf))))

    ;; wrap function call inside start check, if any
    (when function (setf f `(when ,function (funcall ,function ,m ,x))))
    (when start (setf f `(unless (< ,x (the ,d ,start)) ,f)))
    `(let* ((,inf ,info)
            (,fil (midifile-stream ,inf))
            (,s , (coerce 0 d)) (,q  0) (,x ,(coerce 0 d)) ,m)
       (declare (fixnum ,q) (,d ,x ,s) (optimize (speed 3) (safety 0)))
       ,@(if start `((unless (typep ,start ',d)
                       (error "Start ~S not ~S!" ,start ',d))))
       ,@(if end `((unless (typep ,end ',d)
                     (error "End ~S not ~S!" ,end ',d))))
       (loop 
         (multiple-value-setq (,q ,m)
                              (%midifile-read-message ,fil ,inf))
         (when (eq ,m +eot+) (return))
         (setf ,x ,(cond ((eq ttype ':ticks) q)
                         ((eq ttype ':seconds)
                          `(the float (* (the fixnum ,q) 
                                         (the float ,scl))))
                         ((eq ttype ':milliseconds)
                          `(the fixnum
                             (floor 
                              (the float (* (the fixnum ,q) 
                                            (the float ,scl))))))))
         (incf (the ,d ,s) (the ,d ,x))
         ;; splice in the termination clause, if any
         ,@(and end `((when (> ,s (the ,d ,end)) (return))))
         ,@ (and f (list f)))
       ;; if not :eot then we have read one too many messages. reset
       ;; file position back in case the track needs more processing.
       (if (eq ,m +eot+)
         t
         (prog1 nil (%midifile-unread-message ,fil ,m ,q))))))

(defun midifile-map-track (file fn &optional start end)
  (declare (optimize (speed 3) (safety 0)))
  (macrolet ((trackbranch (inf fun typ beg end)
             ;; branch to a different dotrack based on start and end.
             `(if ,beg
                  (if ,end 
                      (dotrack ,inf ,fun ,typ ,beg ,end)
                      (dotrack ,inf ,fun ,typ ,beg nil))
                  (if ,end 
                      (dotrack ,inf ,fun ,typ nil ,end)
                      (dotrack ,inf ,fun ,typ nil nil)))))
    (ecase (mfinfo-ttype file)
      (:ticks (trackbranch file fn :ticks start end))
      (:seconds (trackbranch file fn :seconds start end))
      (:milliseconds (trackbranch file fn :milliseconds start end)))))

(defun midifile-set-track-start (file start)
  (unless (eq (mfinfo-ttype file) ':seconds)
    (error "Only implemented seconds."))
  (setf start (coerce start 'float))
  (unless (= start 0.0)
    (when (dotrack file nil :seconds nil start)
      (error "Start time past end of track ~D in ~S."
        (midifile-track file) (midifile-pathname file)))))

;;;
;;; low level header functions
;;;

(defun %midifile-read-header (stream &aux (bytes (%read-bytes stream 4)))
  (unless (= bytes +MThd+)
    (error "Expected midifile header mark but got ~X instead." bytes))
  (%read-bytes stream 4)            ; skip header length bytes
  (values (%read-bytes stream 2)    ; format
	  (%read-bytes stream 2)    ; ntracks
	  (%read-bytes stream 2)))

(defun %midifile-write-header (stream format tracks division)
  (%write-bytes stream +MThd+ 4)
  (%write-bytes stream 6 4)              ; header length stored 00 00 00 06
  (%write-bytes stream format 2)
  (%write-bytes stream tracks 2)
  (%write-bytes stream division 2)
  stream)

(defun %midifile-read-track-header (stream 
                                     &aux (bytes (%read-bytes stream 4)))
  (unless (= bytes +MTrk+)
    (error "Expected midifile track mark but got ~X instead." bytes))
  (%read-bytes stream 4))

(defun %midifile-write-track-header (stream length)
  (%write-bytes stream +MTrk+ 4)
  (%write-bytes stream length 4)
  stream)

;;;
;;; low level message functions
;;;

(defun %midifile-read-message (stream info)
  ;; returns delta ticks and message, side effects info 
  (let ((ticks 0) (byte 0) (size 0) (raw 0))
    (declare (fixnum byte size raw) (number ticks)
             (inline mfinfo-status mfinfo-size)
             )
    (setf ticks (the fixnum (%read-variable-quantity stream)))
    (setf byte (read-byte stream))
    (cond ((< byte #xf0) ; channel message
           (if (logtest byte 128)
               ;; normal status. cache status byte and message size.
             (setf byte (if (mfinfo-chanmap info)
                          (the fixnum
                            (logior (the fixnum (logand byte #xf0))
                                    (the fixnum
                                      (car
                                       (svref (the (vector t 32) 
                                                (mfinfo-chanmap info))
                                              (the fixnum (logand byte #x0f)))))))
                          byte)
                   raw byte
                   (mfinfo-status info) byte 
                   size (svref (the (vector t 8) +message-sizes+)
                               (ash (logand byte #x70) -4))
                   (mfinfo-size info) size)

               ;; running status. since we skip the status byte, the
               ;; the current byte is already data so we read 1 less
               ;; byte than .size. indicates.
               (setf raw (the fixnum 
                              (logior (the fixnum (ash (mfinfo-status info) 8))
                                      byte))
                     size (1- (the fixnum (mfinfo-size info)))))
           ;; raw = status or status+data if running. read
           ;; any remaining data bytes. use size-1 because size includes
           ;; status byte. 
           (when (> size 1)
             (dotimes (i (1- size))  
               (setf raw (logior (ash raw 8) (read-byte stream)))))
           (values ticks (%midi-encode-message raw (mfinfo-size info))))
          ((= byte +midi-meta+)   ; meta message
           (setf byte (read-byte stream))
           (cond ((= byte +midi-eot+) 
                  (read-byte stream) ; flush the byte
                  (values ticks ':eot))
                 ((= byte +midi-tempo-change+)
                  ;; side effect the tempo file
                  (%read-variable-quantity stream) ; flush length
                  (let ((usecs
                          (logior 
                            (the fixnum
                              (ash (the fixnum (read-byte stream)) 16))
                            (the fixnum
                              (ash (the fixnum (read-byte stream))  8))
                            (the fixnum (read-byte stream)))))
                    (declare (fixnum usecs) )
                    (set-scaler info usecs)
                    (multiple-value-bind (q m)
                       (%midifile-read-message stream info)
                     (values (incf ticks q) m))))
                 (t
                  (setf size (%read-variable-quantity stream))
                  (file-position stream (+ size (file-position stream)))
                  (multiple-value-bind (q m) 
                      (%midifile-read-message stream info)
                    (values (incf ticks q) m)))))
          (t	       ; sysex message ??? system too
           (setf size (%read-variable-quantity stream))
           (file-position stream (+ size (file-position stream)))
           (multiple-value-bind (q m) 
               (%midifile-read-message stream info)
             (values (incf ticks q) m))))))

(defun %midifile-unread-message (stream message &optional (delta 0))
  (file-position stream (- (file-position stream)
                           (+ (%message-size message)
                              (%variable-quantity-length delta)))))

(defun %midifile-write-message (message stream &optional (time 0) data)
  (declare (fixnum time) #-clisp (fixnum message) #+clisp (number message)
           (optimize (speed 3) (safety 0)))
  (let ((size (%message-size message)))
    (declare (fixnum size)  )
    (%write-variable-quantity time stream)
    (write-byte (the fixnum (ash (the fixnum (logand message #xff0000)) -16))
                stream)
    (when (> size 1) 
      (write-byte (the fixnum (ash (the fixnum (logand message #x7f00)) -8))
                  stream))
    (when (> size 2)
      (write-byte (the fixnum (logand message #x7f)) stream))
    (when (= (%message-status message) #xf)  ; system, sysex, meta
      (if data 
        (progn 
          (%write-variable-quantity (length data) stream)
          (dolist (n data)
            (let ((m #xff0000)
                  (s -16))
              (declare (fixnum m s))
              (dotimes (l (%message-size n))
                (write-byte (the fixnum (ash (logand n m) s)) stream)
                (setf m (ash m -8) s (+ s 8))))))
        (%write-variable-quantity 0 stream)))))

(defun %midifile-write-message-data (data stream)
  ;; this is called by stella's after method for midi-message
  (declare (optimize (speed 3) (safety 0)))
  (%write-variable-quantity (length data) stream)
  (dolist (n data)
    (let ((m #xff0000)
          (s -16))
      (declare (fixnum m s))
      (dotimes (l (%message-size n))
        (write-byte (the fixnum (ash (logand n m) s)) stream)
        (setf m (ash m -8) s (+ s 8))))))

;;;
;;; low level byte functions
;;;

(defun %read-variable-quantity (stream &aux (n 0) (b 0))
  (declare (integer n) (fixnum b) (optimize (speed 3) (safety 0)))
  (loop do (setf b (read-byte stream))
           (setf n (the integer (+ (the integer (ash n 7))
                                   (the fixnum (logand b #x7f)))))
        while (logtest #x80 b))
  n)

(defun %write-variable-quantity (n stream)
  (declare (integer n) (optimize (speed 3) (safety 0)))
  (when (>= n 2097152)  ; 1<<21
    (write-byte 
      (the fixnum 
           (logior (the fixnum (logand (the fixnum (ash n -21)) 127)) 128)) 
      stream))
  (when (>= n 16384)    ; 1<<14 
    (write-byte
      (the fixnum
           (logior (the fixnum (logand (the fixnum (ash n -14)) 127)) 128))
      stream))
  (when (>= n 128)      ; 1<<7
    (write-byte
      (the fixnum
           (logior (the fixnum (logand (the fixnum (ash n  -7)) 127)) 128))
    stream))
  (write-byte (the fixnum (logand n 127)) stream))

(defun %variable-quantity-length (n &aux (l 1))
  (declare (integer n) (fixnum l) (optimize (speed 3) (safety 0)))
  (when (>= n 2097152) (incf l))
  (when (>= n 16384) (incf l))
  (when (>= n 128) (incf l))
  l)

(defun %read-bytes (stream n &aux (s 0))
  (declare (fixnum n) (number s) (optimize (speed 3) (safety 0)))
  (loop repeat n do (setf s (+ (ash s 8) (read-byte stream))))
  s)

(defun %write-bytes (stream bytes n)
  (declare (fixnum n) (number bytes) (optimize (speed 3) (safety 0)))
  (loop for i from (* (1- n) 8) downto 0 by 8
        do (write-byte (ldb (byte 8 i) bytes) stream)))

;;;
;;;
;;;

(defun midifile-print (pathname &key (stream *standard-output*) (track 0)
                                     (time-format :ticks))
  (with-open-midifile (file pathname :track track :ttype time-format)
    (let ((*standard-output* stream ))
      (format t "~%File: ~A ~%Format: ~D~%Tracks: ~D~%Division: ~D~
                 ~%~%Track ~D, length ~D"
              (namestring (midifile-pathname file))
              (midifile-format file) (midifile-tracks file)
              (midifile-divisions file)
              (midifile-track file) (midifile-tracklen file))
      (midifile-map-track file #'midi-print-message)
      (values))))

;;;
;;;
;;;

(defun midifile-play (&rest args)
  (if (midi-open?)
    (apply #'process-run-function "MIDI Player" #'midifile-play-aux args)
    (warn-user "MIDI: no port open.")))

;(defvar %midifile-hush nil)

(defun midifile-play-aux (path &key (start 0.0) end (timescale 1.0)
                               (track 0) (headstart 1.0) chanmap mapper
                               (ttype ':seconds)
                               async?
                               &allow-other-keys)
  (let ((scale (* timescale 1000.0))
        time)
    (unless mapper
      (setf mapper 
            (if (not async? )
              #'(lambda (m s) 
                  (%midi-write-message m (round (incf time (* scale s)))))
              #'(lambda (m s &aux a) 
                  (setf a (round (incf time (* scale s))))
                  #+mcl
                  (process-wait "Waiting"
                                #'(lambda (a) (> (midi-get-time) a)) 
                                a)
                  #+excl
                  (loop until (> (midi-get-time) a)
                        do (mp:process-allow-schedule))
                  (%midi-write-message m a)))))
    (with-open-midifile (file path :track track :ttype ttype :chanmap chanmap)
      (when (> start 0.0)
        (midifile-set-track-start file start))
      (when end (setf end (coerce end 'float)))
      (setf time (+ (midi-get-time) (* headstart 1000.0)))
      (unless (catch :hush (midifile-map-track file mapper nil end))
        (%midi-all-notes-off 0)))))
    
;;;
;;; FORMAT T "~%~S ~S"
;;; 

(defmacro midimix (path &rest args)
  `(make-midimix :pathname ,path ,@ args))

(defun midifile-mix (output &rest inputs)
  (let ((queue (make-queue))) 
    (unless (midifile-p output)
      (setf output (make-midifile :pathname output)))
    ;; insure floats, make channel map, if needed
    (dolist (i inputs)
      (when (midimix-start i)
        (setf (midimix-start i) (coerce (midimix-start i) 'float)))
      (when (midimix-end i)
        (setf (midimix-end i) (coerce (midimix-end i) 'float))
        (setf (midimix-hash i) (make-hash-table)))
      (when (midimix-output-start i)
        (setf (midimix-output-start i) 
              (coerce (midimix-output-start i) 'float)))
      (set-ttype i ':seconds)
      (when (midimix-channels i)
        (let ((l (midimix-channels i))
              (m (make-array 16)))
           (if (consp l)
             (dolist (c l)
               (if (consp c)
                   (let ((l (reverse c)))
                     (loop with out = (pop l) for in = (or (pop l) out)
                           do (setf (aref m in) out) until (not l)))
                   (error "Channel map: ~S not list of form: (in ... out)" c)))
             (setf (aref m l) l))
           (setf (midimix-channels i) m))))
    (unwind-protect
      (progn 
        (dolist (i inputs)
          (midifile-open i)
          (when (= (midimix-track i) -1)
            (if (> (midimix-tracks i) 1)
              (error "Not track specification for ~S" (midifile-pathname i))
              (setf (midimix-track i) 0)))
          (midifile-open-track i (midimix-track i) ':seconds)
          (when (> (midimix-start i) 0.0)
            (when (dotrack i nil :seconds nil (midimix-start i))
              (error "Start time past end of track ~D in ~S."
               (midifile-track i) (midifile-pathname i))))
          ;; only enqueue input file if we have a first message.
          (when (midimix-read-message i)
            (when (> (midimix-start i) 0.0)
              (setf (midimix-input-time i) 0.0)
              (setf (midimix-output-time i) (midimix-output-start i)))
            (midimix-enqueue i queue)))
        (when (queue-head queue)
          (midifile-open output :direction ':output)
          (midifile-open-track output 0)
          (let ((s (midifile-stream output)) m d)
            (multiple-value-setq (m d) (make-tempo-change 500000))
            (%midifile-write-message m s 0 d)
            (multiple-value-setq (m d) (make-time-signature 4 4))
            (%midifile-write-message m s 0 d)
            ;; do the mix
            (midimix-output-mix output queue))))
      (midifile-close-track output)
      (midifile-close output)
      (dolist (i inputs) (midifile-close i)))
    (midifile-pathname output)))

(defun midimix-output-mix (output queue)
  ;; process the intput queue and write results the output file. if an input
  ;; ends before +eot+, we check to see if there dangling are note ons. if so,
  ;; a predicate function is computed that is returns t for only missing
  ;; missing note offs and returns +eot+ when complete.
  (let ((time 0.0)
        (last 0.0)
        (scale (/ (mfinfo-divisions output) .5))
        (stream (midifile-stream output))
        in midi diff)
    (loop while (queue-head queue)
          do
          (setf in (midimix-dequeue queue))
          (setf midi (midimix-message in))
          (setf time (midimix-output-time in))
          (setf diff (floor (* (- time last) scale)))
          (%midifile-write-message midi stream diff)
          ;(midi-print-message midi (midimix-output-time in)) 
          ;; requeue if next message, or need note offs.
          (if (midimix-read-message in)
            (midimix-enqueue in queue)
            (unless (eq (midimix-message in) +eot+)
              (let ((fn (midimix-noteoff-predicate in)))
                (when fn
                  (setf (midimix-predicate in) fn)
                  (setf (midimix-end in) nil)
                  (when (midimix-read-message in)
                    (midimix-enqueue in queue))))))
          (setf last time))))

(defun midimix-read-message (file)
  ;; read until next message or eot
  (let ((stream (midifile-stream file))
        (end (midimix-end file))
        (map (midimix-channels file))
        (pred (midimix-predicate file))
        (flag t) m q c s)
    (loop
      (multiple-value-setq (q m) (%midifile-read-message stream file))
      (setf s (* (midimix-scaler file) q))
      (incf (midimix-input-time file) s)
      (when (eq m ':eot) (return (setf flag nil)))
      (when (and end (> (midimix-input-time file) end))
        (return (setf flag nil)))
      (when (setf c (if map (aref map (%message-channel m)) t))    
        (if pred
            (multiple-value-bind (p i) (funcall pred m q)
              (if p (return (and (eq p +eot+) (setf flag nil m p)))
                (if i (decf (midimix-input-time file) s))))
          (return nil)))) 
    (if flag 
        (progn
          (when end
            (let ((type (%message-status m))
                  (tabl (midimix-hash file)))
              (if (= type +midi-note-on+)
                (if (= (%message-velocity m) 0) 
                  (setf (gethash (%message-hash m) tabl)
                        (1- (or (gethash (%message-hash m) tabl) 0)))
                  (setf (gethash (%message-hash m) tabl)
                        (1+ (or (gethash (%message-hash m) tabl) 0))))
                (when (= type +midi-note-off+)
                  (setf (gethash (%message-hash m) tabl)
                        (1- (or (gethash (%message-hash m) tabl) 0)))))))
          (when (midimix-filter file)
            (multiple-value-setq (m s) (funcall (midimix-filter file) m s)))
          (when map 
            (setf m (logior (logand m #xFF0FFFF) (ash c 16)))))
        (unless (eq m ':eot) ; quit before eot so unread last message
         ; (format t "~%flag is ~S unreading" flag)(midi-print-message m)
          (%midifile-unread-message stream m q)
          (decf (midimix-input-time file) s)))
    (setf (midimix-message file) m)
    (setf (midimix-output-time file)
          (+ (midimix-output-start file) (midimix-input-time file)))
    flag))

(defun midimix-noteoff-predicate (in &aux (hash (midimix-hash in)) (count 0))
  ;; if we have dangling note ons, compute a predicate that returns
  ;; true for matching noteoffs.
  (maphash #'(lambda (k v) k (incf count (max v 0))) hash)
  ;(when (> count 0) (format t "~%Searching for ~S offs." count))
  (if (> count 0)
    #'(lambda (m r) r
        ;(format t "~%In drainer with") (midi-print-message m)
        (if (= count 0) ':eot
          (let ((s (%message-status m)))
            (if (or (= s +midi-note-off+)
                    (and (= s +midi-note-on+) (= 0 (%message-velocity m))))
              (if (> (or (gethash (%message-hash m) hash) 0) 0) ; found one
                (progn (decf count) t)
                nil)
              nil))))
    nil))

(defun midimix-dequeue (queue) (pop (queue-head queue)))

(defun midimix-enqueue (mmix queue)
  (if (queue-head queue)
      (cond ((>= (midimix-output-time mmix)
                 (midimix-output-time (car (queue-last queue))))
             (setf (queue-last queue)
                   (cdr (rplacd (queue-last queue) (list mmix)))))
            ((< (midimix-output-time mmix)
                (midimix-output-time (car (queue-head queue))))
             (push mmix (queue-head queue)))	       
            (t 
             (loop with list = (queue-head queue)
                   for head on (cdr list)
                   while (<= (midimix-output-time (car head)) 
                             (midimix-output-time mmix))
                   do (setf list head)
                   finally (rplacd list (cons mmix head)))))
      (progn (setf (queue-head queue) (list mmix))
             (setf (queue-last queue) (queue-head queue))))
  queue)


